from ..Contract import Contract


class FabricaMultisigFactory(Contract):
    def __init__(self, eth_client):
        super().__init__(eth_client, "fabrica-multisig-factory")

    def _check_config(self):
        if self.eth_client.sender_address is None:
            raise ValueError("the sender_address must be not None")

        if self.eth_client.cosigner_address is None:
            raise ValueError("the cosigner_address must be not None")

        if self.eth_client.cosigner_address == self.eth_client.sender_address:
            raise ValueError("the cosigner_address must be different from the sender_address")

    def create(self, customer_address):
        self._check_config()

        customer_address = self.eth_client.w3.toChecksumAddress(customer_address)

        if customer_address in (self.eth_client.sender_address, self.eth_client.cosigner_address):
            raise ValueError("the customer_address must be different from the sender_address and the cosigner_address")

        wallet_address = self.multisig_wallet_address(customer_address)
        if wallet_address is None:
            return self.create_raw([self.eth_client.sender_address, self.eth_client.cosigner_address, customer_address], 2)
        else:
            raise Exception("{} address is already linked to {} wallet".format(customer_address, wallet_address))

    def multisig_wallet_address(self, customer_address):
        result = self.multisig_wallet_address_raw(customer_address)
        if result == "0x0000000000000000000000000000000000000000":
            return None
        else:
            return result

    # AUTO GENERATED #

    def list_end_users(self):
        result = self.contract.functions.listEndUsers().call() 
        return result

    def multisig_wallet_address_raw(self, arg0):
        result = self.contract.functions.multisigWalletAddress(Contract.type_converter(arg0, "address")).call() 
        return result

    def change_address(self, old_address, new_address):
        return self._send_tx(self.contract.functions.changeAddress, **{"oldAddress": Contract.type_converter(old_address, "address"), "newAddress": Contract.type_converter(new_address, "address")}, cosign=False)

    def create_raw(self, owners, required):
        return self._send_tx(self.contract.functions.create, **{"_owners": Contract.type_converter(owners, "address[]"), "_required": Contract.type_converter(required, "uint256")}, cosign=False)



from web3.middleware import geth_poa_middleware
import web3
import hashlib
import ecdsa


class EthereumClient:
    def __init__(self, state, network_url=None, private_key=None, cosigner_address=None):
        self.state = state

        if network_url is None:
            network_url = "http://localhost:8545"

        self.w3 = web3.Web3(web3.HTTPProvider(network_url))
        self.network_id = self.w3.version.network

        if str(self.network_id) in ("4", "42", "77", "99"):
            self.w3.middleware_stack.inject(geth_poa_middleware, layer=0)

        if type(private_key) is str:
            if private_key.startswith("0x"):
                self.private_key = bytes(bytearray.fromhex(private_key[2:]))
            else:
                self.private_key = bytes(bytearray.fromhex(private_key))
        else:
            self.private_key = None

        if self.private_key is None:
            self.sender_address = self.w3.eth.accounts[0]
        else:
            self.sender_address = self.get_address_from_private_key(self.private_key)

        if cosigner_address is None:
            self.cosigner_address = cosigner_address
        else:
            self.cosigner_address = self.w3.toChecksumAddress(cosigner_address)

    def get_nonce(self):
        key = "{}_{}_nonce".format(self.network_id, self.sender_address)
        local_nonce = self.state.get_state(key)
        blockchain_nonce = self.w3.eth.getTransactionCount(self.sender_address)

        if local_nonce is None or blockchain_nonce > local_nonce:
            self.state.set_state(key, blockchain_nonce)
            return blockchain_nonce
        else:
            return local_nonce

    def get_address_from_private_key(self, private_key):
        return self.w3.toChecksumAddress("0x" + self.w3.sha3(hexstr=ecdsa.SigningKey.from_string(string=private_key, curve=ecdsa.SECP256k1, hashfunc=hashlib.sha256).get_verifying_key().to_string().hex())[12:].hex())




import shelve
from os import path
from inspect import isclass


class State:
    def __init__(self, state_plugin=None, local_state_path=None):
        if isclass(state_plugin):
            self.state_plugin = state_plugin()

            for func in "get_state", "set_state":
                if func not in dir(self.state_plugin):
                    raise ValueError("Missing get_state or set_state functions in state_plugin")

            if not callable(self.state_plugin.get_state) or not callable(self.state_plugin.set_state):
                raise ValueError("get_state or set_state in state_plugin are not functions")
        else:
            self.state_plugin = None

            if local_state_path is None:
                self.local_state_path = path.join(path.expanduser("~"), "fabrica-state")
            else:
                self.local_state_path = local_state_path

            self.state = shelve.open(self.local_state_path)

    def set_state(self, key, value):
        if self.state_plugin is not None:
            self.state_plugin.set_state(key, value)
        else:
            self.state[key] = value

    def get_state(self, key):
        if self.state_plugin is not None:
            return self.state_plugin.get_state(key)
        else:
            return self.state.get(key, None)



from ..Contract import Contract
import requests
from rfc3986 import uri_reference


class FabricaToken(Contract):
    def __init__(self, eth_client, ipfs_client, api):
        super().__init__(eth_client, "fabrica-token")
        self.ipfs_client = ipfs_client
        self.api = api

    def _get_token_metadata(self, uri):
        reference = uri_reference(uri)
        scheme = reference.scheme
        if scheme in ("http", "https"):
            try:
                return requests.get(uri).json()
            except:
                # todo: error handling here
                return None
        elif scheme == "ipfs":
            try:
                return self.ipfs_client.get_json(reference.path)
            except:
                # todo: error handling here
                return None
        else:
            return None

    def mint(
        self,
        token_id,
        geohash,
        holding_entity,
        token_metadata,
        cosign=True,
        customer_address=None
    ):
        if self.get_token(token_id) is not None:
            raise ValueError("token {} already exists".format(token_id))

        if isinstance(token_metadata, dict):
            if self.api is None:
                raise ValueError("\"fabrica_api_url\" and \"fabrica_api_key\" must be configured")

            if "document" in token_metadata and isinstance(token_metadata["document"], dict):
                token_metadata["document"] = self.api.create_ipfs_document(token_metadata["document"], "title-holding-trust")

            token_metadata = self.api.create_ipfs_metadata(token_metadata)

        return self.mint_raw(
            token_id=token_id,
            geohash=geohash,
            holding_entity=holding_entity,
            token_uri=token_metadata,
            cosign=cosign,
            customer_address=customer_address
        )

    # todo
    def update_token(self, *args, cosign=True, **kwargs):
        if self.api is not None:
            pass
        else:
            return self.update_token_raw(*args, cosign=cosign, **kwargs)

    def get_token(self, *args, **kwargs):
        token = self.get_token_raw(*args, **kwargs)
        token["metadata"] = self._get_token_metadata(token["token_uri"])
        for key, value in token.items():
            if key != "token_id" and value not in ('', None):
                break
        else:
            return None

        return token

    # AUTO GENERATED #

    def transfer_from(self, from_address, to_address, token_id, cosign=True, customer_address=None):
        return self._send_tx(self.contract.functions.transferFrom, **{"_from": Contract.type_converter(from_address, "address"), "_to": Contract.type_converter(to_address, "address"), "_tokenId": Contract.type_converter(token_id, "uint256")}, cosign=cosign, customer_address=customer_address)

    def mint_raw(self, token_id, geohash, holding_entity, token_uri, cosign=True, customer_address=None):
        return self._send_tx(self.contract.functions.mint, **{"_tokenId": Contract.type_converter(token_id, "uint256"), "_geohash": Contract.type_converter(geohash, "string"), "_holdingEntity": Contract.type_converter(holding_entity, "string"), "_tokenURI": Contract.type_converter(token_uri, "string")}, cosign=cosign, customer_address=customer_address)

    def set_token_uri(self, token_id, token_uri, cosign=True, customer_address=None):
        return self._send_tx(self.contract.functions.setTokenURI, **{"_tokenId": Contract.type_converter(token_id, "uint256"), "_tokenURI": Contract.type_converter(token_uri, "string")}, cosign=cosign, customer_address=customer_address)

    def update_token_raw(self, token_id, holding_entity, token_uri, cosign=True, customer_address=None):
        return self._send_tx(self.contract.functions.updateToken, **{"_tokenId": Contract.type_converter(token_id, "uint256"), "_holdingEntity": Contract.type_converter(holding_entity, "string"), "_tokenURI": Contract.type_converter(token_uri, "string")}, cosign=cosign, customer_address=customer_address)

    def get_token_raw(self, token_id):
        result = self.contract.functions.getToken(**{"_tokenId": Contract.type_converter(token_id, "uint256")}).call() 
        return {"token_id": result[0], "geohash": result[1], "holding_entity": result[2], "token_uri": result[3]}



from posixpath import join
import json
import requests


class FabricaApiClient:
    def __init__(self, fabrica_api_url, fabrica_api_key):
        self.fabrica_api_url = fabrica_api_url
        self.fabrica_api_key = fabrica_api_key

    def _request(self, method, function_name, data=None):
        if isinstance(data, dict):
            data = json.dumps(data)

        r = getattr(requests, method)(
            join(self.fabrica_api_url, function_name),
            headers={
                "x-secret": self.fabrica_api_key,
                "content-type": "application/json"
            },
            data=data
        )

        if r.ok:
            return r.json()
        else:
            raise ConnectionError(r.json())

    def create_ipfs_metadata(self, metadata):
        return self._request("post", "programmatic/createIpfsMetadata", data=metadata)["ipfsUri"]

    def create_ipfs_document(self, data, document_type):
        return self._request("post", "programmatic/createIpfsDocument", data={"type": document_type, "data": data})["ipfsUri"]



import json
import ipfsapi
from os.path import expanduser, isfile, join
from os import linesep
from .EthereumClient import EthereumClient
from .FabricaApiClient import FabricaApiClient
from .State import State
from .contracts.FabricaToken import FabricaToken
from .contracts.FabricaMultisigWallet import FabricaMultisigWallet
from .contracts.FabricaMultisigFactory import FabricaMultisigFactory


class FabricaSdk:
    def __init__(self, config=None, state_plugin=None):
        # config parser
        if config is None:
            default_path = join(expanduser("~"), "fabrica-config.json")
            if isfile(default_path):
                with open(default_path) as json_data_file:
                    self.config = json.load(json_data_file)
            else:
                self.config = {}
        elif isinstance(config, str):
            if isfile(config):
                with open(config) as json_data_file:
                    self.config = json.load(json_data_file)
            else:
                raise Exception("Config file not found")
        elif not isinstance(config, dict):
            raise Exception("Wrong config parameter")

        for key in "private_key", \
                   "cosigner_address", \
                   "fabrica_api_url", \
                   "fabrica_api_key", \
                   "network_url", \
                   "ipfs_endpoint", \
                   "ipfs_port", \
                   "local_state_path":
            if key not in self.config:
                self.config[key] = None

        # state initialization
        self.state = State(
            state_plugin=state_plugin,
            local_state_path=self.config["local_state_path"]
        )

        # ethereum client initialization
        try:
            self.eth_client = EthereumClient(
                self.state,
                network_url=self.config["network_url"],
                private_key=self.config["private_key"],
                cosigner_address=self.config["cosigner_address"]
            )
        except Exception as e:
            self.eth_client = None
            print("WARNING, EthereumClient initialization error:" + linesep + "{}".format(e))

        # api client initialization
        if self.config["fabrica_api_url"] is not None and self.config["fabrica_api_key"] is not None:
            self.api = FabricaApiClient(
                fabrica_api_url=self.config["fabrica_api_url"],
                fabrica_api_key=self.config["fabrica_api_key"]
            )
        else:
            self.api = None

        # ipfs client initialization
        self.ipfs_client = ipfsapi.connect(
            self.config["ipfs_endpoint"] if self.config["ipfs_endpoint"] is not None else "https://ipfs.infura.io",
            self.config["ipfs_port"] if self.config["ipfs_port"] is not None else 5001
        )

        # contracts initialization
        if self.eth_client:
            self.token = self.eth_client.token = FabricaToken(self.eth_client, self.ipfs_client, self.api)
            self.multisig = self.eth_client.multisig = FabricaMultisigWallet(self.eth_client)
            self.factory = self.eth_client.factory = FabricaMultisigFactory(self.eth_client)
        else:
            self.token = self.multisig = self.factory = None



from ..Contract import Contract


class FabricaMultisigWallet(Contract):
    def __init__(self, eth_client):
        super().__init__(eth_client, "fabrica-multisig-wallet")
        self.contracts = {}

    def get_contract(self, customer_address):
        if customer_address not in self.contracts:

            wallet_address = self.eth_client.factory.multisig_wallet_address(customer_address)

            if wallet_address is None:
                raise Exception("no multisig wallets deployed for {}".format(customer_address))

            self.contracts[customer_address] = self.eth_client.w3.eth.contract(address=wallet_address, abi=self.contract_def["abi"])

        return self.contracts[customer_address]

    # AUTO GENERATED #

    def transactions(self, customer_address, arg0):
        result = self.get_contract(customer_address).functions.transactions(Contract.type_converter(arg0, "uint256")).call() 
        return {"destination": result[0], "value": result[1], "data": result[2], "executed": result[3]}

    def get_transaction_count(self, customer_address, pending, executed):
        result = self.get_contract(customer_address).functions.getTransactionCount(**{"pending": Contract.type_converter(pending, "bool"), "executed": Contract.type_converter(executed, "bool")}).call() 
        return {"count": result[0] if isinstance(result, list) else result}

    def get_owners(self, customer_address):
        result = self.get_contract(customer_address).functions.getOwners().call() 
        return result

    def get_transaction_ids(self, customer_address, from_address, to_address, pending, executed):
        result = self.get_contract(customer_address).functions.getTransactionIds(**{"from": Contract.type_converter(from_address, "uint256"), "to": Contract.type_converter(to_address, "uint256"), "pending": Contract.type_converter(pending, "bool"), "executed": Contract.type_converter(executed, "bool")}).call() 
        return {"transaction_ids": result[0] if isinstance(result, list) else result}



from web3 import Web3
import json
import os
from pathlib import Path


class Contract:
    def __init__(self, eth_client, contract_name):
        self.eth_client = eth_client

        # contracts initialization
        self.contract_def = Contract.get_contract_def(contract_name)

        if self.contract_def["networks"] is not None:
            if eth_client.network_id not in self.contract_def["networks"]:
                raise Exception("{} network is not available in {} networks".format(eth_client.network_id, contract_name))

            self.address = eth_client.w3.toChecksumAddress(self.contract_def["networks"][eth_client.network_id]["address"])
            self.contract = eth_client.w3.eth.contract(address=self.address, abi=self.contract_def["abi"])

    def _send_tx(self, contract_function, *args, customer_address=None, cosign=True, **kwargs):
        transaction = contract_function(*args, **kwargs)
        if cosign:
            if customer_address is None:
                print(
                    "To execute {} with a cosigner the positional argument 'customer_address' is required, "
                    "otherwise set the positional argument 'cosign' = False".format(contract_function)
                )
            contract = self.eth_client.multisig.get_contract(customer_address)
            tx_params = {
                "from": self.eth_client.sender_address
            }
            transaction = transaction.buildTransaction(tx_params)
            transaction = contract.functions.submitTransaction(self.address, 0, transaction["data"])

        if self.eth_client.private_key:
            tx_params = {
                "nonce": self.eth_client.get_nonce(),
                "from": self.eth_client.sender_address,
                "value": 0,
                "gas": 4000000,
                "gasPrice": self.eth_client.w3.eth.gasPrice
            }
            transaction = transaction.buildTransaction(tx_params)
            raw_transaction = self.eth_client.w3.eth.account.signTransaction(transaction, self.eth_client.private_key).rawTransaction
            tx_hash = self.eth_client.w3.eth.sendRawTransaction(raw_transaction).hex()
        else:
            tx_params = {
                "from": self.eth_client.sender_address,
                "value": 0,
                "gas": 4000000,
                "gasPrice": self.eth_client.w3.eth.gasPrice
            }
            tx_hash = transaction.transact(tx_params).hex()

        receipt = self.eth_client.w3.eth.waitForTransactionReceipt(tx_hash)
        return receipt

    @staticmethod
    def type_converter(value, required_type):
        if required_type.endswith("[]"):
            return list(map(lambda x: Contract.type_converter(x, required_type.replace("[]", "")), value))
        else:
            if "int" in required_type:
                return Web3.toInt(text=value)
            elif "bytes32" in required_type:
                return Web3.toBytes(text=value).ljust(32, b"\0") \
                    if not value.startswith("0x") else Web3.toBytes(hexstr=value).ljust(32, b"\0")
            elif "byte" in required_type:
                return Web3.toBytes(text=value) \
                    if not value.startswith("0x") else Web3.toBytes(hexstr=value)
            elif "address" in required_type:
                return Web3.toChecksumAddress(value)
            elif "bool" in required_type:
                return bool(value)
            else:
                return str(value)

    @staticmethod
    def get_contract_def(contract_name, contract_artifacts_root=None):
        if contract_artifacts_root is None:
            contract_artifacts_root = Path(__file__).absolute().parent.joinpath("resources", "blockchain")

        contract_def = {}
        with open(Path(__file__).absolute().parent.joinpath(contract_artifacts_root, contract_name, "abi.json")) as f:
            contract_def["abi"] = json.load(f)

        networks_path = Path(__file__).absolute().parent.joinpath(contract_artifacts_root, contract_name, "networks.json")
        if os.path.isfile(networks_path):
            with open(networks_path) as f:
                contract_def["networks"] = json.load(f)
        else:
            contract_def["networks"] = None

        return contract_def



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
   name="fabricasdk",
   version="0.1.1",
   description="fabrica sdk",
   long_description=long_description,
   long_description_content_type="text/markdown",
   author="Fabrica Inc.",
   author_email="info@fabrica.city",
   packages=find_packages(),
   package_data={
       "fabricasdk": ["resources/blockchain/**/*.json", "contracts/*"]
   },
   install_requires=[
       "ecdsa==0.13",
       "ipfsapi==1.11.0",
       "requests==2.19.1",
       "rfc3986==1.1.0",
       "web3==4.6.0"
   ]
)
